import os.path
import pathlib
import tempfile
import unittest
from textwrap import dedent

from rope.base.exceptions import ResourceNotFoundError, RopeError
from rope.base.fscommands import FileSystemCommands
from rope.base.libutils import path_to_resource
from rope.base.project import NoProject, Project, _realpath
from rope.base.resourceobserver import FilteredResourceObserver
from ropetest import testutils


class ProjectTest(unittest.TestCase):
    def setUp(self):
        unittest.TestCase.setUp(self)
        self.project = testutils.sample_project(
            foldername="sampleproject", ropefolder=None
        )
        self.project_root = self.project.address
        self._make_sample_project()
        self.no_project = NoProject()

    def _make_sample_project(self):
        self.sample_file = "sample_file.txt"
        self.sample_path = os.path.join(self.project_root, "sample_file.txt")
        if not os.path.exists(self.project_root):
            os.mkdir(self.project_root)
        self.sample_folder = "sample_folder"
        os.mkdir(os.path.join(self.project_root, self.sample_folder))
        sample = open(self.sample_path, "w")
        sample.write("sample text\n")
        sample.close()

    def tearDown(self):
        testutils.remove_project(self.project)
        unittest.TestCase.tearDown(self)

    def test_project_creation(self):
        self.assertEqual(_realpath(self.project_root), self.project.address)

    def test_getting_project_file(self):
        project_file = self.project.get_resource(self.sample_file)
        self.assertTrue(project_file is not None)

    def test_project_file_reading(self):
        projectFile = self.project.get_resource(self.sample_file)
        self.assertEqual("sample text\n", projectFile.read())

    def test_getting_not_existing_project_file(self):
        with self.assertRaises(ResourceNotFoundError):
            self.project.get_resource("DoesNotExistFile.txt")

    def test_writing_in_project_files(self):
        project_file = self.project.get_resource(self.sample_file)
        project_file.write("another text\n")
        self.assertEqual("another text\n", project_file.read())

    def test_creating_files(self):
        project_file = "newfile.txt"
        self.project.root.create_file(project_file)
        newFile = self.project.get_resource(project_file)
        self.assertTrue(newFile is not None)

    def test_creating_files_that_already_exist(self):
        with self.assertRaises(RopeError):
            self.project.root.create_file(self.sample_file)

    def test_making_root_folder_if_it_does_not_exist(self):
        with tempfile.TemporaryDirectory(dir=testutils.RUN_TMP_DIR) as tmpdir:
            project_root = pathlib.Path(tmpdir) / "sampleproject2"
            assert not project_root.exists()

            project = Project(project_root)

            assert project_root.exists()
            assert project_root.is_dir()

    def test_failure_when_project_root_exists_and_is_a_file(self):
        with tempfile.TemporaryDirectory(dir=testutils.RUN_TMP_DIR) as tmpdir:
            project_root = pathlib.Path(tmpdir) / "sampleproject2"
            project_root.touch()
            assert project_root.exists() and project_root.is_file()

            with self.assertRaises(RopeError):
                Project(project_root)

    def test_creating_folders(self):
        folderName = "SampleFolder"
        self.project.root.create_folder(folderName)
        folderPath = os.path.join(self.project.address, folderName)
        self.assertTrue(os.path.exists(folderPath) and os.path.isdir(folderPath))

    def test_making_folder_that_already_exists(self):
        folderName = "SampleFolder"
        with self.assertRaises(RopeError):
            self.project.root.create_folder(folderName)
            self.project.root.create_folder(folderName)

    def test_failing_if_creating_folder_while_file_already_exists(self):
        folderName = "SampleFolder"
        with self.assertRaises(RopeError):
            self.project.root.create_file(folderName)
            self.project.root.create_folder(folderName)

    def test_creating_file_inside_folder(self):
        folder_name = "sampleFolder"
        file_name = "sample2.txt"
        file_path = folder_name + "/" + file_name
        parent_folder = self.project.root.create_folder(folder_name)
        parent_folder.create_file(file_name)
        file = self.project.get_resource(file_path)
        file.write("sample notes")
        self.assertEqual(file_path, file.path)
        self.assertEqual(
            "sample notes", open(os.path.join(self.project.address, file_path)).read()
        )

    def test_failing_when_creating_file_inside_non_existent_folder(self):
        with self.assertRaises(ResourceNotFoundError):
            self.project.root.create_file("NonexistentFolder/SomeFile.txt")

    def test_nested_directories(self):
        folder_name = "SampleFolder"
        parent = self.project.root.create_folder(folder_name)
        parent.create_folder(folder_name)
        folder_path = os.path.join(self.project.address, folder_name, folder_name)
        self.assertTrue(os.path.exists(folder_path) and os.path.isdir(folder_path))

    def test_removing_files(self):
        self.assertTrue(os.path.exists(self.sample_path))
        self.project.get_resource(self.sample_file).remove()
        self.assertFalse(os.path.exists(self.sample_path))

    def test_removing_files_invalidating_in_project_resource_pool(self):
        root_folder = self.project.root
        my_file = root_folder.create_file("my_file.txt")
        my_file.remove()
        self.assertFalse(root_folder.has_child("my_file.txt"))

    def test_removing_directories(self):
        self.assertTrue(
            os.path.exists(os.path.join(self.project.address, self.sample_folder))
        )
        self.project.get_resource(self.sample_folder).remove()
        self.assertFalse(
            os.path.exists(os.path.join(self.project.address, self.sample_folder))
        )

    def test_removing_non_existent_files(self):
        with self.assertRaises(ResourceNotFoundError):
            self.project.get_resource("NonExistentFile.txt").remove()

    def test_removing_nested_files(self):
        file_name = self.sample_folder + "/sample_file.txt"
        self.project.root.create_file(file_name)
        self.project.get_resource(file_name).remove()
        self.assertTrue(
            os.path.exists(os.path.join(self.project.address, self.sample_folder))
        )
        self.assertTrue(
            not os.path.exists(os.path.join(self.project.address, file_name))
        )

    def test_file_get_name(self):
        file = self.project.get_resource(self.sample_file)
        self.assertEqual(self.sample_file, file.name)
        file_name = "nestedFile.txt"
        parent = self.project.get_resource(self.sample_folder)
        filePath = self.sample_folder + "/" + file_name
        parent.create_file(file_name)
        nestedFile = self.project.get_resource(filePath)
        self.assertEqual(file_name, nestedFile.name)

    def test_folder_get_name(self):
        folder = self.project.get_resource(self.sample_folder)
        self.assertEqual(self.sample_folder, folder.name)

    def test_file_get_path(self):
        file = self.project.get_resource(self.sample_file)
        self.assertEqual(self.sample_file, file.path)
        fileName = "nestedFile.txt"
        parent = self.project.get_resource(self.sample_folder)
        filePath = self.sample_folder + "/" + fileName
        parent.create_file(fileName)
        nestedFile = self.project.get_resource(filePath)
        self.assertEqual(filePath, nestedFile.path)

    def test_folder_get_path(self):
        folder = self.project.get_resource(self.sample_folder)
        self.assertEqual(self.sample_folder, folder.path)

    def test_is_folder(self):
        self.assertTrue(self.project.get_resource(self.sample_folder).is_folder())
        self.assertTrue(not self.project.get_resource(self.sample_file).is_folder())

    def testget_children(self):
        children = self.project.get_resource(self.sample_folder).get_children()
        self.assertEqual([], children)

    def test_nonempty_get_children(self):
        file_name = "nestedfile.txt"
        filePath = self.sample_folder + "/" + file_name
        parent = self.project.get_resource(self.sample_folder)
        parent.create_file(file_name)
        children = parent.get_children()
        self.assertEqual(1, len(children))
        self.assertEqual(filePath, children[0].path)

    def test_nonempty_get_children2(self):
        file_name = "nestedfile.txt"
        folder_name = "nestedfolder.txt"
        filePath = self.sample_folder + "/" + file_name
        folderPath = self.sample_folder + "/" + folder_name
        parent = self.project.get_resource(self.sample_folder)
        parent.create_file(file_name)
        parent.create_folder(folder_name)
        children = parent.get_children()
        self.assertEqual(2, len(children))
        self.assertTrue(filePath == children[0].path or filePath == children[1].path)
        self.assertTrue(
            folderPath == children[0].path or folderPath == children[1].path
        )

    def test_does_not_fail_for_permission_denied(self):
        with tempfile.TemporaryDirectory(suffix="bad_dir") as bad_dir:
            os.chmod(bad_dir, 0o000)
            try:
                parent = self.project.get_resource(self.sample_folder)

                parent.get_children()

            finally:
                os.chmod(bad_dir, 0o755)

    def test_getting_files(self):
        files = self.project.root.get_files()
        self.assertEqual(1, len(files))
        self.assertTrue(self.project.get_resource(self.sample_file) in files)

    def test_getting_folders(self):
        folders = self.project.root.get_folders()
        self.assertEqual(1, len(folders))
        self.assertTrue(self.project.get_resource(self.sample_folder) in folders)

    def test_nested_folder_get_files(self):
        parent = self.project.root.create_folder("top")
        parent.create_file("file1.txt")
        parent.create_file("file2.txt")
        files = parent.get_files()
        self.assertEqual(2, len(files))
        self.assertTrue(self.project.get_resource("top/file2.txt") in files)
        self.assertEqual(0, len(parent.get_folders()))

    def test_nested_folder_get_folders(self):
        parent = self.project.root.create_folder("top")
        parent.create_folder("dir1")
        parent.create_folder("dir2")
        folders = parent.get_folders()
        self.assertEqual(2, len(folders))
        self.assertTrue(self.project.get_resource("top/dir1") in folders)
        self.assertEqual(0, len(parent.get_files()))

    def test_root_folder(self):
        root_folder = self.project.root
        self.assertEqual(2, len(root_folder.get_children()))
        self.assertEqual("", root_folder.path)
        self.assertEqual("", root_folder.name)

    def test_get_all_files(self):
        files = tuple(self.project.get_files())
        self.assertEqual(1, len(files))
        self.assertEqual(self.sample_file, files[0].name)

    def test_get_all_files_after_changing(self):
        self.assertEqual(1, len(self.project.get_files()))
        myfile = self.project.root.create_file("myfile.txt")
        self.assertEqual(2, len(self.project.get_files()))
        myfile.move("newfile.txt")
        self.assertEqual(2, len(self.project.get_files()))
        self.project.get_file("newfile.txt").remove()
        self.assertEqual(1, len(self.project.get_files()))

    def test_multifile_get_all_files(self):
        fileName = "nestedFile.txt"
        parent = self.project.get_resource(self.sample_folder)
        parent.create_file(fileName)
        files = list(self.project.get_files())
        self.assertEqual(2, len(files))
        self.assertTrue(fileName == files[0].name or fileName == files[1].name)

    def test_ignoring_dot_pyc_files_in_get_files(self):
        root = self.project.address
        src_folder = os.path.join(root, "src")
        os.mkdir(src_folder)
        test_pyc = os.path.join(src_folder, "test.pyc")
        open(test_pyc, "w").close()
        for x in self.project.get_files():
            self.assertNotEqual("src/test.pyc", x.path)

    def test_folder_creating_files(self):
        projectFile = "NewFile.txt"
        self.project.root.create_file(projectFile)
        new_file = self.project.get_resource(projectFile)
        self.assertTrue(new_file is not None and not new_file.is_folder())

    def test_folder_creating_nested_files(self):
        project_file = "NewFile.txt"
        parent_folder = self.project.get_resource(self.sample_folder)
        parent_folder.create_file(project_file)
        new_file = self.project.get_resource(self.sample_folder + "/" + project_file)
        self.assertTrue(new_file is not None and not new_file.is_folder())

    def test_folder_creating_files2(self):
        projectFile = "newfolder"
        self.project.root.create_folder(projectFile)
        new_folder = self.project.get_resource(projectFile)
        self.assertTrue(new_folder is not None and new_folder.is_folder())

    def test_folder_creating_nested_files2(self):
        project_file = "newfolder"
        parent_folder = self.project.get_resource(self.sample_folder)
        parent_folder.create_folder(project_file)
        new_folder = self.project.get_resource(self.sample_folder + "/" + project_file)
        self.assertTrue(new_folder is not None and new_folder.is_folder())

    def test_folder_get_child(self):
        folder = self.project.root
        folder.create_file("myfile.txt")
        folder.create_folder("myfolder")
        self.assertEqual(
            self.project.get_resource("myfile.txt"), folder.get_child("myfile.txt")
        )
        self.assertEqual(
            self.project.get_resource("myfolder"), folder.get_child("myfolder")
        )

    def test_folder_get_child_nested(self):
        root = self.project.root
        folder = root.create_folder("myfolder")
        folder.create_file("myfile.txt")
        folder.create_folder("myfolder")
        self.assertEqual(
            self.project.get_resource("myfolder/myfile.txt"),
            folder.get_child("myfile.txt"),
        )
        self.assertEqual(
            self.project.get_resource("myfolder/myfolder"), folder.get_child("myfolder")
        )

    def test_project_root_is_root_folder(self):
        self.assertEqual("", self.project.root.path)

    def test_moving_files(self):
        root_folder = self.project.root
        my_file = root_folder.create_file("my_file.txt")
        my_file.move("my_other_file.txt")
        self.assertFalse(my_file.exists())
        root_folder.get_child("my_other_file.txt")

    def test_moving_folders(self):
        root_folder = self.project.root
        my_folder = root_folder.create_folder("my_folder")
        my_file = my_folder.create_file("my_file.txt")
        my_folder.move("new_folder")
        self.assertFalse(root_folder.has_child("my_folder"))
        self.assertFalse(my_file.exists())
        self.assertTrue(root_folder.get_child("new_folder") is not None)

    def test_moving_destination_folders(self):
        root_folder = self.project.root
        my_folder = root_folder.create_folder("my_folder")
        my_file = root_folder.create_file("my_file.txt")
        my_file.move("my_folder")
        self.assertFalse(root_folder.has_child("my_file.txt"))
        self.assertFalse(my_file.exists())
        my_folder.get_child("my_file.txt")

    def test_moving_files_and_resource_objects(self):
        root_folder = self.project.root
        my_file = root_folder.create_file("my_file.txt")
        old_hash = hash(my_file)
        my_file.move("my_other_file.txt")
        self.assertEqual(old_hash, hash(my_file))

    def test_file_encoding_reading(self):
        sample_file = self.project.root.create_file("my_file.txt")
        contents = (
            b"# -*- coding: utf-8 -*-\n"
            + br"#\N{LATIN SMALL LETTER I WITH DIAERESIS}\n"
        ).decode("utf8")
        file = open(sample_file.real_path, "wb")
        file.write(contents.encode("utf-8"))
        file.close()
        self.assertEqual(contents, sample_file.read())

    def test_file_encoding_writing(self):
        sample_file = self.project.root.create_file("my_file.txt")
        contents = (
            b"# -*- coding: utf-8 -*-\n" + br"\N{LATIN SMALL LETTER I WITH DIAERESIS}\n"
        ).decode("utf8")
        sample_file.write(contents)
        self.assertEqual(contents, sample_file.read())

    def test_using_utf8_when_writing_in_case_of_errors(self):
        sample_file = self.project.root.create_file("my_file.txt")
        contents = br"\n\N{LATIN SMALL LETTER I WITH DIAERESIS}\n".decode("utf8")
        sample_file.write(contents)
        self.assertEqual(contents, sample_file.read())

    def test_encoding_declaration_in_the_second_line(self):
        sample_file = self.project.root.create_file("my_file.txt")
        contents = b"\n# -*- coding: latin-1 -*-\n\xa9\n"
        file = open(sample_file.real_path, "wb")
        file.write(contents)
        file.close()
        self.assertEqual(contents, sample_file.read().encode("latin-1"))

    def test_not_an_encoding_declaration(self):
        sample_file = self.project.root.create_file("my_file.txt")
        contents = b"def my_method(self, encoding='latin-1'):\n    var = {}\n\xc2\xa9\n"
        file = open(sample_file.real_path, "wb")
        file.write(contents)
        file.close()
        self.assertEqual(contents, sample_file.read().encode("utf-8"))
        self.assertNotEqual(contents, sample_file.read().encode("latin-1"))

    def test_read_bytes(self):
        sample_file = self.project.root.create_file("my_file.txt")
        contents = b"\n# -*- coding: latin-1 -*-\n\xa9\n"
        file = open(sample_file.real_path, "wb")
        file.write(contents)
        file.close()
        self.assertEqual(contents, sample_file.read_bytes())

    def test_file_with_unix_line_ending(self):
        sample_file = self.project.root.create_file("my_file.txt")
        contents = b"1\n"
        file = open(sample_file.real_path, "wb")
        file.write(contents)
        file.close()
        self.assertIsNone(sample_file.newlines)
        self.assertEqual("1\n", sample_file.read())
        self.assertEqual("\n", sample_file.newlines)

        sample_file.write("1\n")
        self.assertEqual(b"1\n", sample_file.read_bytes())

    def test_file_with_dos_line_ending(self):
        sample_file = self.project.root.create_file("my_file.txt")
        contents = b"1\r\n"
        file = open(sample_file.real_path, "wb")
        file.write(contents)
        file.close()
        self.assertIsNone(sample_file.newlines)
        self.assertEqual("1\n", sample_file.read())
        self.assertEqual("\r\n", sample_file.newlines)

        sample_file.write("1\n")
        self.assertEqual(b"1\r\n", sample_file.read_bytes())

    def test_file_with_mac_line_ending(self):
        sample_file = self.project.root.create_file("my_file.txt")
        contents = b"1\r"
        file = open(sample_file.real_path, "wb")
        file.write(contents)
        file.close()
        self.assertIsNone(sample_file.newlines)
        self.assertEqual("1\n", sample_file.read())
        self.assertEqual("\r", sample_file.newlines)

        sample_file.write("1\n")
        self.assertEqual(b"1\r", sample_file.read_bytes())

    def test_file_binary(self):
        sample_file = self.project.root.create_file("my_file.txt")
        contents = b"1\r\n"
        file = open(sample_file.real_path, "wb")
        file.write(contents)
        file.close()
        self.assertIsNone(sample_file.newlines)
        self.assertEqual(b"1\r\n", sample_file.read_bytes())
        self.assertIsNone(sample_file.newlines)

        sample_file.write(b"1\nx\r")
        self.assertEqual((b"1\nx\r"), sample_file.read_bytes())

    # TODO: Detecting utf-16 encoding
    def xxx_test_using_utf16(self):
        sample_file = self.project.root.create_file("my_file.txt")
        contents = b"# -*- coding: utf-16 -*-\n# This is a sample file ...\n"
        file = open(sample_file.real_path, "w")
        file.write(contents.encode("utf-16"))
        file.close()
        sample_file.write(contents)
        self.assertEqual(contents, sample_file.read())

    # XXX: supporting utf_8_sig
    def xxx_test_file_encoding_reading_for_notepad_styles(self):
        sample_file = self.project.root.create_file("my_file.txt")
        contents = "#\N{LATIN SMALL LETTER I WITH DIAERESIS}\n"
        file = open(sample_file.real_path, "w")
        # file.write('\xef\xbb\xbf')
        file.write(contents.encode("utf-8-sig"))
        file.close()
        self.assertEqual(contents, sample_file.read())

    def test_using_project_get_file(self):
        myfile = self.project.get_file(self.sample_file)
        self.assertTrue(myfile.exists())

    def test_using_file_create(self):
        myfile = self.project.get_file("myfile.txt")
        self.assertFalse(myfile.exists())
        myfile.create()
        self.assertTrue(myfile.exists())
        self.assertFalse(myfile.is_folder())

    def test_using_folder_create(self):
        myfolder = self.project.get_folder("myfolder")
        self.assertFalse(myfolder.exists())
        myfolder.create()
        self.assertTrue(myfolder.exists())
        self.assertTrue(myfolder.is_folder())

    def test_exception_when_creating_twice(self):
        with self.assertRaises(RopeError):
            myfile = self.project.get_file("myfile.txt")
            myfile.create()
            myfile.create()

    def test_exception_when_parent_does_not_exist(self):
        with self.assertRaises(ResourceNotFoundError):
            myfile = self.project.get_file("myfolder/myfile.txt")
            myfile.create()

    def test_simple_path_to_resource(self):
        myfile = self.project.root.create_file("myfile.txt")
        self.assertEqual(myfile, path_to_resource(self.project, myfile.real_path))
        self.assertEqual(
            myfile, path_to_resource(self.project, myfile.real_path, type="file")
        )
        myfolder = self.project.root.create_folder("myfolder")
        self.assertEqual(myfolder, path_to_resource(self.project, myfolder.real_path))
        self.assertEqual(
            myfolder, path_to_resource(self.project, myfolder.real_path, type="folder")
        )

    @testutils.skipNotPOSIX()
    def test_ignoring_symlinks_inside_project(self):
        project2 = testutils.sample_project(folder_name="sampleproject2")
        mod = project2.root.create_file("mod.py")
        try:
            path = os.path.join(self.project.address, "linkedfile.txt")
            os.symlink(mod.real_path, path)
            files = self.project.root.get_files()
            self.assertEqual(1, len(files))
        finally:
            testutils.remove_project(project2)

    def test_getting_empty_source_folders(self):
        self.assertEqual([], self.project.get_source_folders())

    def test_root_source_folder(self):
        self.project.root.create_file("sample.py")
        source_folders = self.project.get_source_folders()
        self.assertEqual(1, len(source_folders))
        self.assertTrue(self.project.root in source_folders)

    def test_root_source_folder2(self):
        self.project.root.create_file("mod1.py")
        self.project.root.create_file("mod2.py")
        source_folders = self.project.get_source_folders()
        self.assertEqual(1, len(source_folders))
        self.assertTrue(self.project.root in source_folders)

    def test_src_source_folder(self):
        src = self.project.root.create_folder("src")
        src.create_file("sample.py")
        source_folders = self.project.get_source_folders()
        self.assertEqual(1, len(source_folders))
        self.assertTrue(self.project.get_resource("src") in source_folders)

    def test_packages(self):
        src = self.project.root.create_folder("src")
        pkg = src.create_folder("package")
        pkg.create_file("__init__.py")
        source_folders = self.project.get_source_folders()
        self.assertEqual(1, len(source_folders))
        self.assertTrue(src in source_folders)

    def test_multi_source_folders(self):
        src = self.project.root.create_folder("src")
        package = src.create_folder("package")
        package.create_file("__init__.py")
        test = self.project.root.create_folder("test")
        test.create_file("alltests.py")
        source_folders = self.project.get_source_folders()
        self.assertEqual(2, len(source_folders))
        self.assertTrue(src in source_folders)
        self.assertTrue(test in source_folders)

    def test_multi_source_folders2(self):
        testutils.create_module(self.project, "mod1")
        src = self.project.root.create_folder("src")
        package = testutils.create_package(self.project, "package", src)
        testutils.create_module(self.project, "mod2", package)
        source_folders = self.project.get_source_folders()
        self.assertEqual(2, len(source_folders))
        self.assertTrue(self.project.root in source_folders and src in source_folders)


class ResourceObserverTest(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.project = testutils.sample_project()

    def tearDown(self):
        testutils.remove_project(self.project)
        super().tearDown()

    def test_resource_change_observer(self):
        sample_file = self.project.root.create_file("my_file.txt")
        sample_file.write("a sample file version 1")
        sample_observer = _SampleObserver()
        self.project.add_observer(sample_observer)
        sample_file.write("a sample file version 2")
        self.assertEqual(1, sample_observer.change_count)
        self.assertEqual(sample_file, sample_observer.last_changed)

    def test_resource_change_observer_after_removal(self):
        sample_file = self.project.root.create_file("my_file.txt")
        sample_file.write("text")
        sample_observer = _SampleObserver()
        self.project.add_observer(
            FilteredResourceObserver(sample_observer, [sample_file])
        )
        sample_file.remove()
        self.assertEqual(1, sample_observer.change_count)
        self.assertEqual(sample_file, sample_observer.last_removed)

    def test_resource_change_observer2(self):
        sample_file = self.project.root.create_file("my_file.txt")
        sample_observer = _SampleObserver()
        self.project.add_observer(sample_observer)
        self.project.remove_observer(sample_observer)
        sample_file.write("a sample file version 2")
        self.assertEqual(0, sample_observer.change_count)

    def test_resource_change_observer_for_folders(self):
        root_folder = self.project.root
        my_folder = root_folder.create_folder("my_folder")
        my_folder_observer = _SampleObserver()
        root_folder_observer = _SampleObserver()
        self.project.add_observer(
            FilteredResourceObserver(my_folder_observer, [my_folder])
        )
        self.project.add_observer(
            FilteredResourceObserver(root_folder_observer, [root_folder])
        )
        my_file = my_folder.create_file("my_file.txt")
        self.assertEqual(1, my_folder_observer.change_count)
        my_file.move("another_file.txt")
        self.assertEqual(2, my_folder_observer.change_count)
        self.assertEqual(1, root_folder_observer.change_count)
        self.project.get_resource("another_file.txt").remove()
        self.assertEqual(2, my_folder_observer.change_count)
        self.assertEqual(2, root_folder_observer.change_count)

    def test_resource_change_observer_after_moving(self):
        sample_file = self.project.root.create_file("my_file.txt")
        sample_observer = _SampleObserver()
        self.project.add_observer(sample_observer)
        sample_file.move("new_file.txt")
        self.assertEqual(1, sample_observer.change_count)
        self.assertEqual(
            (sample_file, self.project.get_resource("new_file.txt")),
            sample_observer.last_moved,
        )

    def test_revalidating_files(self):
        root = self.project.root
        my_file = root.create_file("my_file.txt")
        sample_observer = _SampleObserver()
        self.project.add_observer(FilteredResourceObserver(sample_observer, [my_file]))
        os.remove(my_file.real_path)
        self.project.validate(root)
        self.assertEqual(my_file, sample_observer.last_removed)
        self.assertEqual(1, sample_observer.change_count)

    def test_revalidating_files_and_no_changes2(self):
        root = self.project.root
        my_file = root.create_file("my_file.txt")
        sample_observer = _SampleObserver()
        self.project.add_observer(FilteredResourceObserver(sample_observer, [my_file]))
        self.project.validate(root)
        self.assertEqual(None, sample_observer.last_moved)
        self.assertEqual(0, sample_observer.change_count)

    def test_revalidating_folders(self):
        root = self.project.root
        my_folder = root.create_folder("myfolder")
        my_file = my_folder.create_file("myfile.txt")  # noqa
        sample_observer = _SampleObserver()
        self.project.add_observer(
            FilteredResourceObserver(sample_observer, [my_folder])
        )
        testutils.remove_recursively(my_folder.real_path)
        self.project.validate(root)
        self.assertEqual(my_folder, sample_observer.last_removed)
        self.assertEqual(1, sample_observer.change_count)

    def test_removing_and_adding_resources_to_filtered_observer(self):
        my_file = self.project.root.create_file("my_file.txt")
        sample_observer = _SampleObserver()
        filtered_observer = FilteredResourceObserver(sample_observer)
        self.project.add_observer(filtered_observer)
        my_file.write("1")
        self.assertEqual(0, sample_observer.change_count)
        filtered_observer.add_resource(my_file)
        my_file.write("2")
        self.assertEqual(1, sample_observer.change_count)
        filtered_observer.remove_resource(my_file)
        my_file.write("3")
        self.assertEqual(1, sample_observer.change_count)

    def test_validation_and_changing_files(self):
        my_file = self.project.root.create_file("my_file.txt")
        sample_observer = _SampleObserver()
        timekeeper = _MockChangeIndicator()
        filtered_observer = FilteredResourceObserver(
            sample_observer, [my_file], timekeeper=timekeeper
        )
        self.project.add_observer(filtered_observer)
        self._write_file(my_file.real_path)
        timekeeper.set_indicator(my_file, 1)
        self.project.validate(self.project.root)
        self.assertEqual(1, sample_observer.change_count)

    def test_validation_and_changing_files2(self):
        my_file = self.project.root.create_file("my_file.txt")
        sample_observer = _SampleObserver()
        timekeeper = _MockChangeIndicator()
        self.project.add_observer(
            FilteredResourceObserver(sample_observer, [my_file], timekeeper=timekeeper)
        )
        timekeeper.set_indicator(my_file, 1)
        my_file.write("hey")
        self.assertEqual(1, sample_observer.change_count)
        self.project.validate(self.project.root)
        self.assertEqual(1, sample_observer.change_count)

    def test_not_reporting_multiple_changes_to_folders(self):
        root = self.project.root
        file1 = root.create_file("file1.txt")
        file2 = root.create_file("file2.txt")
        sample_observer = _SampleObserver()
        self.project.add_observer(
            FilteredResourceObserver(sample_observer, [root, file1, file2])
        )
        os.remove(file1.real_path)
        os.remove(file2.real_path)
        self.assertEqual(0, sample_observer.change_count)
        self.project.validate(self.project.root)
        self.assertEqual(3, sample_observer.change_count)

    def _write_file(self, path):
        my_file = open(path, "w")
        my_file.write("\n")
        my_file.close()

    def test_moving_and_being_interested_about_a_folder_and_a_child(self):
        my_folder = self.project.root.create_folder("my_folder")
        my_file = my_folder.create_file("my_file.txt")
        sample_observer = _SampleObserver()
        filtered_observer = FilteredResourceObserver(
            sample_observer, [my_folder, my_file]
        )
        self.project.add_observer(filtered_observer)
        my_folder.move("new_folder")
        self.assertEqual(2, sample_observer.change_count)

    def test_contains_for_folders(self):
        folder1 = self.project.root.create_folder("folder")
        folder2 = self.project.root.create_folder("folder2")
        self.assertFalse(folder1.contains(folder2))

    def test_validating_when_created(self):
        root = self.project.root
        my_file = self.project.get_file("my_file.txt")
        sample_observer = _SampleObserver()
        self.project.add_observer(FilteredResourceObserver(sample_observer, [my_file]))
        open(my_file.real_path, "w").close()
        self.project.validate(root)
        self.assertEqual(my_file, sample_observer.last_created)
        self.assertEqual(1, sample_observer.change_count)

    def test_validating_twice_when_created(self):
        root = self.project.root
        my_file = self.project.get_file("my_file.txt")
        sample_observer = _SampleObserver()
        self.project.add_observer(FilteredResourceObserver(sample_observer, [my_file]))
        open(my_file.real_path, "w").close()
        self.project.validate(root)
        self.project.validate(root)
        self.assertEqual(my_file, sample_observer.last_created)
        self.assertEqual(1, sample_observer.change_count)

    def test_changes_and_adding_resources(self):
        root = self.project.root  # noqa
        file1 = self.project.get_file("file1.txt")
        file2 = self.project.get_file("file2.txt")
        file1.create()
        sample_observer = _SampleObserver()
        self.project.add_observer(
            FilteredResourceObserver(sample_observer, [file1, file2])
        )
        file1.move(file2.path)
        self.assertEqual(2, sample_observer.change_count)
        self.assertEqual(file2, sample_observer.last_created)
        self.assertEqual((file1, file2), sample_observer.last_moved)

    def test_validating_get_files_list(self):
        root = self.project.root  # noqa
        self.assertEqual(0, len(self.project.get_files()))
        file = open(os.path.join(self.project.address, "myfile.txt"), "w")
        file.close()
        self.project.validate()
        self.assertEqual(1, len(self.project.get_files()))

    def test_clear_observed_resources_for_filtered_observers(self):
        sample_file = self.project.root.create_file("myfile.txt")
        sample_observer = _SampleObserver()
        filtered = FilteredResourceObserver(sample_observer)
        self.project.add_observer(filtered)
        filtered.add_resource(sample_file)
        filtered.clear_resources()
        sample_file.write("1")
        self.assertEqual(0, sample_observer.change_count)


class _MockChangeIndicator:
    def __init__(self):
        self.times = {}

    def set_indicator(self, resource, time):
        self.times[resource] = time

    def get_indicator(self, resource):
        return self.times.get(resource, 0)


class _SampleObserver:
    def __init__(self):
        self.change_count = 0
        self.last_changed = None
        self.last_moved = None
        self.last_created = None
        self.last_removed = None

    def resource_changed(self, resource):
        self.last_changed = resource
        self.change_count += 1

    def resource_moved(self, resource, new_resource):
        self.last_moved = (resource, new_resource)
        self.change_count += 1

    def resource_created(self, resource):
        self.last_created = resource
        self.change_count += 1

    def resource_removed(self, resource):
        self.last_removed = resource
        self.change_count += 1


class OutOfProjectTest(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.test_directory = "temp_test_directory"
        testutils.remove_recursively(self.test_directory)
        os.mkdir(self.test_directory)
        self.project = testutils.sample_project()
        self.no_project = NoProject()

    def tearDown(self):
        testutils.remove_project(self.project)
        testutils.remove_recursively(self.test_directory)
        super().tearDown()

    def test_simple_out_of_project_file(self):
        sample_file_path = os.path.join(self.test_directory, "sample.txt")
        sample_file = open(sample_file_path, "w")
        sample_file.write("sample content\n")
        sample_file.close()
        sample_resource = self.no_project.get_resource(sample_file_path)
        self.assertEqual("sample content\n", sample_resource.read())

    def test_simple_out_of_project_folder(self):
        sample_folder_path = os.path.join(self.test_directory, "sample_folder")
        os.mkdir(sample_folder_path)
        sample_folder = self.no_project.get_resource(sample_folder_path)
        self.assertEqual([], sample_folder.get_children())

        sample_file_path = os.path.join(sample_folder_path, "sample.txt")
        open(sample_file_path, "w").close()
        sample_resource = self.no_project.get_resource(sample_file_path)
        self.assertEqual(sample_resource, sample_folder.get_children()[0])

    def test_using_absolute_path(self):
        sample_file_path = os.path.join(self.test_directory, "sample.txt")
        open(sample_file_path, "w").close()
        normal_sample_resource = self.no_project.get_resource(sample_file_path)
        absolute_sample_resource = self.no_project.get_resource(
            os.path.abspath(sample_file_path)
        )
        self.assertEqual(normal_sample_resource, absolute_sample_resource)

    def test_folder_get_child(self):
        sample_folder_path = os.path.join(self.test_directory, "sample_folder")
        os.mkdir(sample_folder_path)
        sample_folder = self.no_project.get_resource(sample_folder_path)
        self.assertEqual([], sample_folder.get_children())

        sample_file_path = os.path.join(sample_folder_path, "sample.txt")
        open(sample_file_path, "w").close()
        sample_resource = self.no_project.get_resource(sample_file_path)
        self.assertTrue(sample_folder.has_child("sample.txt"))
        self.assertFalse(sample_folder.has_child("doesnothave.txt"))
        self.assertEqual(sample_resource, sample_folder.get_child("sample.txt"))

    def test_out_of_project_files_and_path_to_resource(self):
        sample_file_path = os.path.join(self.test_directory, "sample.txt")
        sample_file = open(sample_file_path, "w")
        sample_file.write("sample content\n")
        sample_file.close()
        sample_resource = self.no_project.get_resource(sample_file_path)
        self.assertEqual(
            sample_resource, path_to_resource(self.project, sample_file_path)
        )


class _MockFSCommands:
    def __init__(self):
        self.log = ""
        self.fscommands = FileSystemCommands()

    def create_file(self, path):
        self.log += "create_file "
        self.fscommands.create_file(path)

    def create_folder(self, path):
        self.log += "create_folder "
        self.fscommands.create_folder(path)

    def move(self, path, new_location):
        self.log += "move "
        self.fscommands.move(path, new_location)

    def remove(self, path):
        self.log += "remove "
        self.fscommands.remove(path)

    def read(self, path):
        self.log += "read "
        return self.fscommands.read(path)


class _DeprecatedFSCommands:
    def __init__(self):
        self.log = ""
        self.fscommands = FileSystemCommands()

    def create_file(self, path):
        self.log += "create_file "
        self.fscommands.create_file(path)

    def create_folder(self, path):
        self.log += "create_folder "
        self.fscommands.create_folder(path)

    def move(self, path, new_location):
        self.log += "move "
        self.fscommands.move(path, new_location)

    def remove(self, path):
        self.log += "remove "
        self.fscommands.remove(path)


class RopeFolderTest(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.project = None

    def tearDown(self):
        if self.project:
            testutils.remove_project(self.project)
        super().tearDown()

    def test_none_project_rope_folder(self):
        self.project = testutils.sample_project(ropefolder=None)
        self.assertTrue(self.project.ropefolder is None)

    def test_getting_project_rope_folder(self):
        self.project = testutils.sample_project()
        self.assertTrue(self.project.ropefolder.exists())
        self.assertEqual(".ropeproject", self.project.ropefolder.path)

    def test_setting_ignored_resources(self):
        self.project = testutils.sample_project(ignored_resources=["myfile.txt"])
        myfile = self.project.get_file("myfile.txt")
        file2 = self.project.get_file("file2.txt")
        self.assertTrue(self.project.is_ignored(myfile))
        self.assertFalse(self.project.is_ignored(file2))

    def test_ignored_folders(self):
        self.project = testutils.sample_project(ignored_resources=["myfolder"])
        myfolder = self.project.root.create_folder("myfolder")
        self.assertTrue(self.project.is_ignored(myfolder))
        myfile = myfolder.create_file("myfile.txt")
        self.assertTrue(self.project.is_ignored(myfile))

    def test_ignored_resources_and_get_files(self):
        self.project = testutils.sample_project(
            ignored_resources=["myfile.txt"], ropefolder=None
        )
        myfile = self.project.get_file("myfile.txt")
        self.assertEqual(0, len(self.project.get_files()))
        myfile.create()
        self.assertEqual(0, len(self.project.get_files()))

    def test_ignored_resources_and_get_files2(self):
        self.project = testutils.sample_project(
            ignored_resources=["myfile.txt"], ropefolder=None
        )
        myfile = self.project.root.create_file("myfile.txt")  # noqa
        self.assertEqual(0, len(self.project.get_files()))

    def test_setting_ignored_resources_patterns(self):
        self.project = testutils.sample_project(ignored_resources=["m?file.*"])
        myfile = self.project.get_file("myfile.txt")
        file2 = self.project.get_file("file2.txt")
        self.assertTrue(self.project.is_ignored(myfile))
        self.assertFalse(self.project.is_ignored(file2))

    def test_star_should_not_include_slashes(self):
        self.project = testutils.sample_project(ignored_resources=["f*.txt"])
        folder = self.project.root.create_folder("folder")
        file1 = folder.create_file("myfile.txt")
        file2 = folder.create_file("file2.txt")
        self.assertFalse(self.project.is_ignored(file1))
        self.assertTrue(self.project.is_ignored(file2))

    def test_normal_fscommands(self):
        fscommands = _MockFSCommands()
        self.project = testutils.sample_project(fscommands=fscommands)
        myfile = self.project.get_file("myfile.txt")
        myfile.create()
        self.assertEqual("create_file ", fscommands.log)

    def test_fscommands_and_ignored_resources(self):
        fscommands = _MockFSCommands()
        self.project = testutils.sample_project(
            fscommands=fscommands, ignored_resources=["myfile.txt"], ropefolder=None
        )
        myfile = self.project.get_file("myfile.txt")
        myfile.create()
        self.assertEqual("", fscommands.log)

    def test_deprecated_fscommands(self):
        fscommands = _DeprecatedFSCommands()
        self.project = testutils.sample_project(fscommands=fscommands)
        myfile = self.project.get_file("myfile.txt")
        myfile.create()
        self.assertEqual("create_file ", fscommands.log)

    def test_ignored_resources_and_prefixes(self):
        self.project = testutils.sample_project(ignored_resources=[".hg"])
        myfile = self.project.root.create_file(".hgignore")
        self.assertFalse(self.project.is_ignored(myfile))

    def test_loading_config_dot_py(self):
        self.project = testutils.sample_project()
        config = self.project.get_file(".ropeproject/config.py")
        if not config.exists():
            config.create()
        config.write(dedent("""\
            def set_prefs(prefs):
                prefs["ignored_resources"] = ["myfile.txt"]
            def project_opened(project):
                project.root.create_file("loaded")
        """))
        self.project.close()
        self.project = Project(self.project.address)
        self.assertTrue(self.project.get_file("loaded").exists())
        myfile = self.project.get_file("myfile.txt")
        self.assertTrue(self.project.is_ignored(myfile))

    def test_loading_pyproject(self):
        self.project = testutils.sample_project()
        config = self.project.get_file("pyproject.toml")
        if not config.exists():
            config.create()
        config.write(dedent("""\
            [tool.rope]
            ignored_resources=["pyproject.py"]
        """))
        self.project.close()
        self.project = Project(self.project.address)
        myfile = self.project.get_file("pyproject.py")
        self.assertTrue(self.project.is_ignored(myfile))

    def test_loading_pyproject_empty_file(self):
        self.project = testutils.sample_project()
        config = self.project.get_file("pyproject.toml")
        if not config.exists():
            config.create()
        config.write("")
        self.project.close()
        self.project = Project(self.project.address)
        myfile = self.project.get_file("pyproject.py")
        self.assertFalse(self.project.is_ignored(myfile))

    def test_loading_pyproject_no_tool_section(self):
        self.project = testutils.sample_project()
        config = self.project.get_file("pyproject.toml")
        if not config.exists():
            config.create()
        config.write(dedent("""\
            [project]
            name = 'testproject'
        """))
        self.project.close()
        self.project = Project(self.project.address)
        myfile = self.project.get_file("pyproject.py")
        self.assertFalse(self.project.is_ignored(myfile))

    def test_loading_pyproject_no_tool_rope_section(self):
        self.project = testutils.sample_project()
        config = self.project.get_file("pyproject.toml")
        if not config.exists():
            config.create()
        config.write(dedent("""\
            [tool.anothertool]
            name = 'testproject'
        """))
        self.project.close()
        self.project = Project(self.project.address)
        myfile = self.project.get_file("pyproject.py")
        self.assertFalse(self.project.is_ignored(myfile))

    def test_ignoring_syntax_errors(self):
        self.project = testutils.sample_project(
            ropefolder=None, ignore_syntax_errors=True
        )
        mod = testutils.create_module(self.project, "mod")
        mod.write("xyz print")
        pymod = self.project.get_pymodule(mod)  # noqa

    def test_compressed_history(self):
        self.project = testutils.sample_project(compress_history=True)
        mod = testutils.create_module(self.project, "mod")
        mod.write("")

    def test_compressed_objectdb(self):
        self.project = testutils.sample_project(compress_objectdb=True)
        mod = testutils.create_module(self.project, "mod")
        self.project.pycore.analyze_module(mod)

    def test_nested_dot_ropeproject_folder(self):
        self.project = testutils.sample_project(ropefolder=".f1/f2")
        ropefolder = self.project.ropefolder
        self.assertEqual(".f1/f2", ropefolder.path)
        self.assertTrue(ropefolder.exists())



"""A module to ease code analysis

This module is here to help source code analysis.
"""
import re

from rope.base import codeanalyze, utils


@utils.cached(7)
def real_code(source):
    """Simplify `source` for analysis

    It replaces:

    * comments with spaces
    * strs with a new str filled with spaces
    * implicit and explicit continuations with spaces
    * tabs and semicolons with spaces

    The resulting code is a lot easier to analyze if we are interested
    only in offsets.
    """
    collector = codeanalyze.ChangeCollector(source)
    for start, end, matchgroups in ignored_regions(source):
        if source[start] == "#":
            replacement = " " * (end - start)
        elif "f" in matchgroups.get("prefix", "").lower():
            replacement = None
        else:
            replacement = '"%s"' % (" " * (end - start - 2))
        if replacement is not None:
            collector.add_change(start, end, replacement)
    source = collector.get_changed() or source
    collector = codeanalyze.ChangeCollector(source)
    parens = 0
    for match in _parens.finditer(source):
        i = match.start()
        c = match.group()
        if c in "({[":
            parens += 1
        if c in ")}]":
            parens -= 1
        if c == "\n" and parens > 0:
            collector.add_change(i, i + 1, " ")
    source = collector.get_changed() or source
    return source.replace("\\\n", "  ").replace("\t", " ").replace(";", "\n")


@utils.cached(7)
def ignored_regions(source):
    """Return ignored regions like strings and comments in `source`"""
    return [
        (match.start(), match.end(), match.groupdict())
        for match in _str.finditer(source)
    ]


_str = re.compile(
    "|".join(
        [
            codeanalyze.get_comment_pattern(),
            codeanalyze.get_any_string_pattern(),
        ]
    )
)
_parens = re.compile(r"[\({\[\]}\)\n]")



# Based on super lightweight Simple Top-Down Parser from http://effbot.org/zone/simple-top-down-parsing.htm
# and https://bitbucket.org/emacsway/sqlbuilder/src/default/sqlbuilder/smartsql/contrib/evaluate.py
import re

from rope.base import utils as base_utils
from rope.base.oi.type_hinting import utils


class SymbolBase:

    name = None  # node/token type name

    def __init__(self):
        self.value = None  # used by name and literals
        self.first = None
        self.second = None
        self.third = None  # used by tree nodes

    def nud(self, parser):
        raise SyntaxError("Syntax error (%r)." % self.name)

    def led(self, left, parser):
        raise SyntaxError("Unknown operator (%r)." % self.name)

    def evaluate(self, pyobject):
        raise NotImplementedError(self.name, self)

    def __repr__(self):
        if self.name == "(name)":
            return f"({self.name[1:-1]} {self.value})"
        out = [repr(self.name), self.first, self.second, self.third]
        out = [str(i) for i in out if i]
        return "(" + " ".join(out) + ")"


class SymbolTable:
    def multi(func):
        def _inner(self, names, *a, **kw):
            for name in names.split():
                func(self, name, *a, **kw)

        return _inner

    def __init__(self):
        self.symbol_table = {}

    def get(self, name, default=None):
        return self.symbol_table.get(name, default)

    def __getitem__(self, name):
        return self.symbol_table[name]

    def __iter__(self):
        return iter(self.symbol_table)

    def symbol(self, name, bp=0):
        try:
            s = self.symbol_table[name]
        except KeyError:

            class S(SymbolBase):
                pass

            s = S
            s.__name__ = "symbol-" + name  # for debugging
            s.name = name
            s.lbp = bp
            self.symbol_table[name] = s
        else:
            s.lbp = max(bp, s.lbp)
        return s

    @multi
    def infix(self, name, bp):
        symbol = self.symbol(name, bp)

        @method(symbol)
        def led(self, left, parser):
            self.first = left
            self.second = parser.expression(bp)
            return self

    @multi
    def infix_r(self, name, bp):
        symbol = self.symbol(name, bp)

        @method(symbol)
        def led(self, left, parser):
            self.first = left
            self.second = parser.expression(bp - 0.1)
            return self

    def ternary(self, name, name2, bp):
        symbol = self.symbol(name, bp)
        symbol2 = self.symbol(name2)

        @method(symbol)
        def led(self, left, parser):
            self.first = left
            self.second = parser.expression(symbol2.lbp)
            parser.advance(symbol2.name)
            self.third = parser.expression(symbol2.lbp + 0.1)
            return self

    @multi
    def prefix(self, name, bp):
        symbol = self.symbol(name, bp)

        @method(symbol)
        def nud(self, parser):
            self.first = parser.expression(bp)
            return self

    @multi
    def postfix(self, name, bp):
        symbol = self.symbol(name, bp)

        @method(symbol)
        def led(self, left, parser):
            self.first = left
            return self

    multi = staticmethod(multi)  # Just for code checker


symbol_table = SymbolTable()


class Lexer:

    _token_pattern = re.compile(
        r"""
        \s*
        (?:
              (
                    [,()\[\]|]
                  | ->
                  | (?<=\s)(?:or)\b
              )  # operator
            | ([a-zA-Z](?:\w|\.)*)  # name
        )
        """,
        re.U | re.S | re.X,
    )

    def __init__(self, symbol_table):
        self.symbol_table = symbol_table

    def tokenize(self, program):
        for name, value in self._tokenize_expr(program):
            symbol = symbol_table.get(value)
            if symbol:
                s = symbol()
            elif name == "(name)":
                symbol = symbol_table[name]
                s = symbol()
                s.value = value
            else:
                raise SyntaxError(
                    "Unknown operator ({}). Possible operators are {!r}".format(
                        value, list(self.symbol_table)
                    )
                )

            yield s

    def _tokenize_expr(self, program):
        if isinstance(program, bytes):
            program = program.decode("utf-8")
        # import pprint; pprint.pprint(self._token_pattern.findall(program))
        for operator, name in self._token_pattern.findall(program):
            if operator:
                yield "(operator)", operator
            elif name:
                yield "(name)", name
            else:
                raise SyntaxError
        yield "(end)", "(end)"


class Parser:

    token = None
    next = None

    def __init__(self, lexer):
        self.lexer = lexer

    def parse(self, program):
        generator = self.lexer.tokenize(program)
        self.next = generator.__next__
        self.token = self.next()
        return self.expression()

    def expression(self, rbp=0):
        t = self.token
        self.token = self.next()
        left = t.nud(self)
        while rbp < self.token.lbp:
            t = self.token
            self.token = self.next()
            left = t.led(left, self)
        return left

    def advance(self, name=None):
        if name and self.token.name != name:
            raise SyntaxError(f"Expected {name!r} but found {self.token.name!r}")
        self.token = self.next()


def method(s):
    assert issubclass(s, SymbolBase)

    def bind(fn):
        setattr(s, fn.__name__, fn)
        return fn

    return bind


symbol, infix, infix_r, prefix, postfix, ternary = (
    symbol_table.symbol,
    symbol_table.infix,
    symbol_table.infix_r,
    symbol_table.prefix,
    symbol_table.postfix,
    symbol_table.ternary,
)

symbol("(", 270)
symbol(")")
symbol("[", 250)  # Parameters
symbol("]")
symbol("->", 230)
infix("|", 170)
infix("or", 170)
symbol(",")

symbol("(name)")
symbol("(end)")


@method(symbol("(name)"))
def nud(self, parser):
    return self


@method(symbol("(name)"))
def evaluate(self, pyobject):
    return utils.resolve_type(self.value, pyobject)


# Parametrized objects
@method(symbol("["))
def led(self, left, parser):
    self.first = left
    self.second = []
    if parser.token.name != "]":
        while 1:
            if parser.token.name == "]":
                break
            self.second.append(parser.expression())
            if parser.token.name != ",":
                break
            parser.advance(",")
    parser.advance("]")
    return self


@method(symbol("["))
def evaluate(self, pyobject):
    return utils.parametrize_type(
        self.first.evaluate(pyobject), *[i.evaluate(pyobject) for i in self.second]
    )


# Anonymous Function Calls
@method(symbol("("))
def nud(self, parser):
    self.second = []
    if parser.token.name != ")":
        while 1:
            self.second.append(parser.expression())
            if parser.token.name != ",":
                break
            parser.advance(",")
    parser.advance(")")
    parser.advance("->")
    self.third = parser.expression(symbol("->").lbp + 0.1)
    return self


# Function Calls
@method(symbol("("))
def led(self, left, parser):
    self.first = left
    self.second = []
    if parser.token.name != ")":
        while 1:
            self.second.append(parser.expression())
            if parser.token.name != ",":
                break
            parser.advance(",")
    parser.advance(")")
    parser.advance("->")
    self.third = parser.expression(symbol("->").lbp + 0.1)
    return self


@method(symbol("("))
def evaluate(self, pyobject):
    # TODO: Implement me
    raise NotImplementedError


@method(symbol("or"))
@method(symbol("|"))
def evaluate(self, pyobject):
    # TODO: Implement me
    raise NotImplementedError


class Compiler:

    parser_factory = Parser
    lexer_factory = Lexer
    symbol_table = symbol_table

    def _make_parser(self):
        return self.parser_factory(self.lexer_factory(self.symbol_table))

    @base_utils.cached(500)
    def __call__(self, program):
        """
        :type program: str
        :rtype: rope.base.oi.type_hinting.evaluate.SymbolBase
        """
        return self._make_parser().parse(program)


compile = Compiler()


class Evaluator:

    compile = compile

    def __call__(self, program, pyobject):
        """Evaluates the program string or AST

        :type program: str or rope.base.oi.type_hinting.evaluate.SymbolBase
        :rtype: rope.base.pyobjects.PyDefinedObject | rope.base.pyobjects.PyObject or None
        """
        ast = self.compile(program) if isinstance(program, str) else program
        return ast.evaluate(pyobject)


evaluate = Evaluator()



import keyword
import sys
import warnings

from rope.base import (
    builtins,
    evaluate,
    exceptions,
    libutils,
    pynames,
    pynamesdef,
    pyobjects,
    pyobjectsdef,
    pyscopes,
    worder,
)
from rope.contrib import fixsyntax
from rope.refactor import functionutils


def code_assist(
    project,
    source_code,
    offset,
    resource=None,
    templates=None,
    maxfixes=1,
    later_locals=True,
):
    """Return python code completions as a list of `CodeAssistProposal`

    `resource` is a `rope.base.resources.Resource` object.  If
    provided, relative imports are handled.

    `maxfixes` is the maximum number of errors to fix if the code has
    errors in it.

    If `later_locals` is `False` names defined in this scope and after
    this line is ignored.

    """
    if templates is not None:
        warnings.warn(
            "Codeassist no longer supports templates", DeprecationWarning, stacklevel=2
        )
    assist = _PythonCodeAssist(
        project,
        source_code,
        offset,
        resource=resource,
        maxfixes=maxfixes,
        later_locals=later_locals,
    )
    return assist()


def starting_offset(source_code, offset):
    """Return the offset in which the completion should be inserted

    Usually code assist proposals should be inserted like::

        completion = proposal.name
        result = (source_code[:starting_offset] +
                  completion + source_code[offset:])

    Where starting_offset is the offset returned by this function.

    """
    word_finder = worder.Worder(source_code, True)
    expression, starting, starting_offset = word_finder.get_splitted_primary_before(
        offset
    )
    return starting_offset


def get_doc(project, source_code, offset, resource=None, maxfixes=1):
    """Get the pydoc"""
    fixer = fixsyntax.FixSyntax(project, source_code, resource, maxfixes)
    pyname = fixer.pyname_at(offset)
    if pyname is None:
        return None
    pyobject = pyname.get_object()
    return PyDocExtractor().get_doc(pyobject)


def get_calltip(
    project,
    source_code,
    offset,
    resource=None,
    maxfixes=1,
    ignore_unknown=False,
    remove_self=False,
):
    """Get the calltip of a function

    The format of the returned string is
    ``module_name.holding_scope_names.function_name(arguments)``.  For
    classes `__init__()` and for normal objects `__call__()` function
    is used.

    Note that the offset is on the function itself *not* after the its
    open parenthesis.  (Actually it used to be the other way but it
    was easily confused when string literals were involved.  So I
    decided it is better for it not to try to be too clever when it
    cannot be clever enough).  You can use a simple search like::

        offset = source_code.rindex('(', 0, offset) - 1

    to handle simple situations.

    If `ignore_unknown` is `True`, `None` is returned for functions
    without source-code like builtins and extensions.

    If `remove_self` is `True`, the first parameter whose name is self
    will be removed for methods.
    """
    fixer = fixsyntax.FixSyntax(project, source_code, resource, maxfixes)
    pyname = fixer.pyname_at(offset)
    if pyname is None:
        return None
    pyobject = pyname.get_object()
    return PyDocExtractor().get_calltip(pyobject, ignore_unknown, remove_self)


def get_definition_location(project, source_code, offset, resource=None, maxfixes=1):
    """Return the definition location of the python name at `offset`

    Return a (`rope.base.resources.Resource`, lineno) tuple.  If no
    `resource` is given and the definition is inside the same module,
    the first element of the returned tuple would be `None`.  If the
    location cannot be determined ``(None, None)`` is returned.

    """
    fixer = fixsyntax.FixSyntax(project, source_code, resource, maxfixes)
    pyname = fixer.pyname_at(offset)
    if pyname is not None:
        module, lineno = pyname.get_definition_location()
        if module is not None:
            return module.get_module().get_resource(), lineno
    return (None, None)


def find_occurrences(*args, **kwds):
    import rope.contrib.findit

    warnings.warn(
        "Use `rope.contrib.findit.find_occurrences()` instead",
        DeprecationWarning,
        stacklevel=2,
    )
    return rope.contrib.findit.find_occurrences(*args, **kwds)


def get_canonical_path(project, resource, offset):
    """Get the canonical path to an object.

    Given the offset of the object, this returns a list of
    (name, name_type) tuples representing the canonical path to the
    object. For example, the 'x' in the following code:

        class Foo(object):
            def bar(self):
                class Qux(object):
                    def mux(self, x):
                        pass

    we will return:

        [('Foo', 'CLASS'), ('bar', 'FUNCTION'), ('Qux', 'CLASS'),
         ('mux', 'FUNCTION'), ('x', 'PARAMETER')]

    `resource` is a `rope.base.resources.Resource` object.

    `offset` is the offset of the pyname you want the path to.

    """
    # Retrieve the PyName.
    pymod = project.get_pymodule(resource)
    pyname = evaluate.eval_location(pymod, offset)

    # Now get the location of the definition and its containing scope.
    defmod, lineno = pyname.get_definition_location()
    if not defmod:
        return None
    scope = defmod.get_scope().get_inner_scope_for_line(lineno)

    # Start with the name of the object we're interested in.
    names = []
    if isinstance(pyname, pynamesdef.ParameterName):
        names = [(worder.get_name_at(pymod.get_resource(), offset), "PARAMETER")]
    elif isinstance(pyname, pynamesdef.AssignedName):
        names = [(worder.get_name_at(pymod.get_resource(), offset), "VARIABLE")]

    # Collect scope names.
    while scope.parent:
        if isinstance(scope, pyscopes.FunctionScope):
            scope_type = "FUNCTION"
        elif isinstance(scope, pyscopes.ClassScope):
            scope_type = "CLASS"
        else:
            scope_type = None
        names.append((scope.pyobject.get_name(), scope_type))
        scope = scope.parent

    names.append((defmod.get_resource().real_path, "MODULE"))
    names.reverse()
    return names


class CompletionProposal:
    """A completion proposal

    The `scope` instance variable shows where proposed name came from
    and can be 'global', 'local', 'builtin', 'attribute', 'keyword',
    'imported', 'parameter_keyword'.

    The `type` instance variable shows the approximate type of the
    proposed object and can be 'instance', 'class', 'function', 'module',
    and `None`.

    All possible relations between proposal's `scope` and `type` are shown
    in the table below (different scopes in rows and types in columns):

                      | instance | class | function | module | None
                local |    +     |   +   |    +     |   +    |
               global |    +     |   +   |    +     |   +    |
              builtin |    +     |   +   |    +     |        |
            attribute |    +     |   +   |    +     |   +    |
             imported |    +     |   +   |    +     |   +    |
              keyword |          |       |          |        |  +
    parameter_keyword |          |       |          |        |  +

    """

    def __init__(self, name, scope, pyname=None):
        self.name = name
        self.pyname = pyname
        self.scope = self._get_scope(scope)

    def __str__(self):
        return f"{self.name} ({self.scope}, {self.type})"

    def __repr__(self):
        return str(self)

    @property
    def parameters(self):
        """The names of the parameters the function takes.

        Returns None if this completion is not a function.
        """
        pyname = self.pyname
        if isinstance(pyname, pynames.ImportedName):
            pyname = pyname._get_imported_pyname()
        if isinstance(pyname, pynames.DefinedName):
            pyobject = pyname.get_object()
            if isinstance(pyobject, pyobjects.AbstractFunction):
                return pyobject.get_param_names()

    @property
    def type(self):
        pyname = self.pyname
        if isinstance(pyname, builtins.BuiltinName):
            pyobject = pyname.get_object()
            if isinstance(pyobject, builtins.BuiltinFunction):
                return "function"
            elif isinstance(pyobject, builtins.BuiltinClass):
                return "class"
            elif isinstance(pyobject, builtins.BuiltinObject) or isinstance(
                pyobject, builtins.BuiltinName
            ):
                return "instance"
        elif isinstance(pyname, pynames.ImportedModule):
            return "module"
        elif isinstance(pyname, pynames.ImportedName) or isinstance(
            pyname, pynames.DefinedName
        ):
            pyobject = pyname.get_object()
            if isinstance(pyobject, pyobjects.AbstractFunction):
                return "function"
            if isinstance(pyobject, pyobjects.AbstractClass):
                return "class"
        return "instance"

    def _get_scope(self, scope):
        if isinstance(self.pyname, builtins.BuiltinName):
            return "builtin"
        if isinstance(self.pyname, pynames.ImportedModule) or isinstance(
            self.pyname, pynames.ImportedName
        ):
            return "imported"
        return scope

    def get_doc(self):
        """Get the proposed object's docstring.

        Returns None if it can not be get.
        """
        if not self.pyname:
            return None
        pyobject = self.pyname.get_object()
        if not hasattr(pyobject, "get_doc"):
            return None
        return self.pyname.get_object().get_doc()

    @property
    def kind(self):
        warnings.warn(
            "the proposal's `kind` property is deprecated, " "use `scope` instead"
        )
        return self.scope


# leaved for backward compatibility
CodeAssistProposal = CompletionProposal


class NamedParamProposal(CompletionProposal):
    """A parameter keyword completion proposal

    Holds reference to ``_function`` -- the function which
    parameter ``name`` belongs to. This allows to determine
    default value for this parameter.
    """

    def __init__(self, name, function):
        self.argname = name
        name = "%s=" % name
        super().__init__(name, "parameter_keyword")
        self._function = function

    def get_default(self):
        """Get a string representation of a param's default value.

        Returns None if there is no default value for this param.
        """
        definfo = functionutils.DefinitionInfo.read(self._function)
        for arg, default in definfo.args_with_defaults:
            if self.argname == arg:
                return default
        return None


def sorted_proposals(proposals, scopepref=None, typepref=None):
    """Sort a list of proposals

    Return a sorted list of the given `CodeAssistProposal`.

    `scopepref` can be a list of proposal scopes.  Defaults to
    ``['parameter_keyword', 'local', 'global', 'imported',
    'attribute', 'builtin', 'keyword']``.

    `typepref` can be a list of proposal types.  Defaults to
    ``['class', 'function', 'instance', 'module', None]``.
    (`None` stands for completions with no type like keywords.)
    """
    sorter = _ProposalSorter(proposals, scopepref, typepref)
    return sorter.get_sorted_proposal_list()


def starting_expression(source_code, offset):
    """Return the expression to complete"""
    word_finder = worder.Worder(source_code, True)
    expression, starting, starting_offset = word_finder.get_splitted_primary_before(
        offset
    )
    if expression:
        return expression + "." + starting
    return starting


def default_templates():
    warnings.warn(
        "default_templates() is deprecated.", DeprecationWarning, stacklevel=2
    )
    return {}


class _PythonCodeAssist:
    def __init__(
        self, project, source_code, offset, resource=None, maxfixes=1, later_locals=True
    ):
        self.project = project
        self.code = source_code
        self.resource = resource
        self.maxfixes = maxfixes
        self.later_locals = later_locals
        self.word_finder = worder.Worder(source_code, True)
        (
            self.expression,
            self.starting,
            self.offset,
        ) = self.word_finder.get_splitted_primary_before(offset)

    keywords = keyword.kwlist

    def _find_starting_offset(self, source_code, offset):
        current_offset = offset - 1
        while current_offset >= 0 and (
            source_code[current_offset].isalnum() or source_code[current_offset] in "_"
        ):
            current_offset -= 1
        return current_offset + 1

    def _matching_keywords(self, starting):
        return [
            CompletionProposal(kw, "keyword")
            for kw in self.keywords
            if kw.startswith(starting)
        ]

    def __call__(self):
        if self.offset > len(self.code):
            return []
        completions = list(self._code_completions().values())
        if self.expression.strip() == "" and self.starting.strip() != "":
            completions.extend(self._matching_keywords(self.starting))
        return completions

    def _dotted_completions(self, module_scope, holding_scope):
        result = {}
        found_pyname = evaluate.eval_str(holding_scope, self.expression)
        if found_pyname is not None:
            element = found_pyname.get_object()
            compl_scope = "attribute"
            if isinstance(element, (pyobjectsdef.PyModule, pyobjectsdef.PyPackage)):
                compl_scope = "imported"
            for name, pyname in element.get_attributes().items():
                if name.startswith(self.starting):
                    result[name] = CompletionProposal(name, compl_scope, pyname)
        return result

    def _undotted_completions(self, scope, result, lineno=None):
        if scope.parent is not None:
            self._undotted_completions(scope.parent, result)
        if lineno is None:
            names = scope.get_propagated_names()
        else:
            names = scope.get_names()
        for name, pyname in names.items():
            if name.startswith(self.starting):
                compl_scope = "local"
                if scope.get_kind() == "Module":
                    compl_scope = "global"
                if (
                    lineno is None
                    or self.later_locals
                    or not self._is_defined_after(scope, pyname, lineno)
                ):
                    result[name] = CompletionProposal(name, compl_scope, pyname)

    def _from_import_completions(self, pymodule):
        module_name = self.word_finder.get_from_module(self.offset)
        if module_name is None:
            return {}
        pymodule = self._find_module(pymodule, module_name)
        result = {}
        for name in pymodule:
            if name.startswith(self.starting):
                result[name] = CompletionProposal(
                    name, scope="global", pyname=pymodule[name]
                )
        return result

    def _find_module(self, pymodule, module_name):
        dots = 0
        while module_name[dots] == ".":
            dots += 1
        pyname = pynames.ImportedModule(pymodule, module_name[dots:], dots)
        return pyname.get_object()

    def _is_defined_after(self, scope, pyname, lineno):
        location = pyname.get_definition_location()
        if location is not None and location[1] is not None:
            if (
                location[0] == scope.pyobject.get_module()
                and lineno <= location[1] <= scope.get_end()
            ):
                return True

    def _code_completions(self):
        lineno = self.code.count("\n", 0, self.offset) + 1
        fixer = fixsyntax.FixSyntax(
            self.project, self.code, self.resource, self.maxfixes
        )
        pymodule = fixer.get_pymodule()
        module_scope = pymodule.get_scope()
        code = pymodule.source_code
        lines = code.split("\n")
        result = {}
        start = fixsyntax._logical_start(lines, lineno)
        indents = fixsyntax._get_line_indents(lines[start - 1])
        inner_scope = module_scope.get_inner_scope_for_line(start, indents)
        if self.word_finder.is_a_name_after_from_import(self.offset):
            return self._from_import_completions(pymodule)
        if self.expression.strip() != "":
            result.update(self._dotted_completions(module_scope, inner_scope))
        else:
            result.update(self._keyword_parameters(module_scope.pyobject, inner_scope))
            self._undotted_completions(inner_scope, result, lineno=lineno)
        return result

    def _keyword_parameters(self, pymodule, scope):
        offset = self.offset
        if offset == 0:
            return {}
        word_finder = worder.Worder(self.code, True)
        if word_finder.is_on_function_call_keyword(offset - 1):
            function_parens = word_finder.find_parens_start_from_inside(offset - 1)
            primary = word_finder.get_primary_at(function_parens - 1)
            try:
                function_pyname = evaluate.eval_str(scope, primary)
            except exceptions.BadIdentifierError:
                return {}
            if function_pyname is not None:
                pyobject = function_pyname.get_object()
                if isinstance(pyobject, pyobjects.AbstractFunction):
                    pass
                elif (
                    isinstance(pyobject, pyobjects.AbstractClass)
                    and "__init__" in pyobject
                ):
                    pyobject = pyobject["__init__"].get_object()
                elif "__call__" in pyobject:
                    pyobject = pyobject["__call__"].get_object()
                if isinstance(pyobject, pyobjects.AbstractFunction):
                    param_names = []
                    param_names.extend(pyobject.get_param_names(special_args=False))
                    result = {}
                    for name in param_names:
                        if name.startswith(self.starting):
                            result[name + "="] = NamedParamProposal(name, pyobject)
                    return result
        return {}


class _ProposalSorter:
    """Sort a list of code assist proposals"""

    def __init__(self, code_assist_proposals, scopepref=None, typepref=None):
        self.proposals = code_assist_proposals
        if scopepref is None:
            scopepref = [
                "parameter_keyword",
                "local",
                "global",
                "imported",
                "attribute",
                "builtin",
                "keyword",
            ]
        self.scopepref = scopepref
        if typepref is None:
            typepref = ["class", "function", "instance", "module", None]
        self.typerank = {type: index for index, type in enumerate(typepref)}

    def get_sorted_proposal_list(self):
        """Return a list of `CodeAssistProposal`"""
        proposals = {}
        for proposal in self.proposals:
            proposals.setdefault(proposal.scope, []).append(proposal)
        result = []
        for scope in self.scopepref:
            scope_proposals = proposals.get(scope, [])
            scope_proposals = [
                proposal
                for proposal in scope_proposals
                if proposal.type in self.typerank
            ]
            scope_proposals.sort(key=self._proposal_key)
            result.extend(scope_proposals)
        return result

    def _proposal_key(self, proposal1):
        def _underline_count(name):
            return sum(1 for c in name if c == "_")

        return (
            self.typerank.get(proposal1.type, 100),
            _underline_count(proposal1.name),
            proposal1.name,
        )
        # if proposal1.type != proposal2.type:
        #    return cmp(self.typerank.get(proposal1.type, 100),
        #               self.typerank.get(proposal2.type, 100))
        # return self._compare_underlined_names(proposal1.name,
        #                                      proposal2.name)


class PyDocExtractor:
    def get_doc(self, pyobject):
        if isinstance(pyobject, pyobjects.AbstractFunction):
            return self._get_function_docstring(pyobject)
        elif isinstance(pyobject, pyobjects.AbstractClass):
            return self._get_class_docstring(pyobject)
        elif isinstance(pyobject, pyobjects.AbstractModule):
            return self._trim_docstring(pyobject.get_doc())
        return None

    def get_calltip(self, pyobject, ignore_unknown=False, remove_self=False):
        try:
            if isinstance(pyobject, pyobjects.AbstractClass):
                pyobject = pyobject["__init__"].get_object()
            if not isinstance(pyobject, pyobjects.AbstractFunction):
                pyobject = pyobject["__call__"].get_object()
        except exceptions.AttributeNotFoundError:
            return None
        if ignore_unknown and not isinstance(pyobject, pyobjects.PyFunction):
            return
        if isinstance(pyobject, pyobjects.AbstractFunction):
            result = self._get_function_signature(pyobject, add_module=True)
            if remove_self and self._is_method(pyobject):
                return result.replace("(self)", "()").replace("(self, ", "(")
            return result

    def _get_class_docstring(self, pyclass):
        def _get_class_header(pyclass):
            class_name = pyclass.get_name()

            supers = [super.get_name() for super in pyclass.get_superclasses()]
            super_classes = ", ".join(supers)

            return f"class {class_name}({super_classes}):\n\n"

        contents = self._trim_docstring(pyclass.get_doc(), 2)
        doc = _get_class_header(pyclass)
        doc += contents

        if "__init__" in pyclass:
            init = pyclass["__init__"].get_object()
            if isinstance(init, pyobjects.AbstractFunction):
                doc += "\n\n" + self._get_single_function_docstring(init)
        return doc

    def _get_function_docstring(self, pyfunction):
        functions = [pyfunction]
        if self._is_method(pyfunction):
            functions.extend(
                self._get_super_methods(pyfunction.parent, pyfunction.get_name())
            )
        return "\n\n".join(
            [self._get_single_function_docstring(function) for function in functions]
        )

    def _is_method(self, pyfunction):
        return isinstance(pyfunction, pyobjects.PyFunction) and isinstance(
            pyfunction.parent, pyobjects.PyClass
        )

    def _get_single_function_docstring(self, pyfunction):
        signature = self._get_function_signature(pyfunction)
        docs = self._trim_docstring(pyfunction.get_doc(), indents=2)
        return signature + ":\n\n" + docs

    def _get_super_methods(self, pyclass, name):
        result = []
        for super_class in pyclass.get_superclasses():
            if name in super_class:
                function = super_class[name].get_object()
                if isinstance(function, pyobjects.AbstractFunction):
                    result.append(function)
            result.extend(self._get_super_methods(super_class, name))
        return result

    def _get_function_signature(self, pyfunction, add_module=False):
        location = self._location(pyfunction, add_module)
        if isinstance(pyfunction, pyobjects.PyFunction):
            info = functionutils.DefinitionInfo.read(pyfunction)
            return location + info.to_string()
        else:
            return "{}({})".format(
                location + pyfunction.get_name(),
                ", ".join(pyfunction.get_param_names()),
            )

    def _location(self, pyobject, add_module=False):
        location = []
        parent = pyobject.parent
        while parent and not isinstance(parent, pyobjects.AbstractModule):
            location.append(parent.get_name())
            location.append(".")
            parent = parent.parent
        if add_module:
            if isinstance(pyobject, pyobjects.PyFunction):
                location.insert(0, self._get_module(pyobject))
            if isinstance(parent, builtins.BuiltinModule):
                location.insert(0, parent.get_name() + ".")
        return "".join(location)

    def _get_module(self, pyfunction):
        module = pyfunction.get_module()
        if module is not None:
            resource = module.get_resource()
            if resource is not None:
                return libutils.modname(resource) + "."
        return ""

    def _trim_docstring(self, docstring, indents=0):
        """The sample code from :PEP:`257`"""
        if not docstring:
            return ""
        # Convert tabs to spaces (following normal Python rules)
        # and split into a list of lines:
        lines = docstring.expandtabs().splitlines()
        # Determine minimum indentation (first line doesn't count):
        indent = sys.maxsize
        for line in lines[1:]:
            stripped = line.lstrip()
            if stripped:
                indent = min(indent, len(line) - len(stripped))
        # Remove indentation (first line is special):
        trimmed = [lines[0].strip()]
        if indent < sys.maxsize:
            for line in lines[1:]:
                trimmed.append(line[indent:].rstrip())
        # Strip off trailing and leading blank lines:
        while trimmed and not trimmed[-1]:
            trimmed.pop()
        while trimmed and not trimmed[0]:
            trimmed.pop(0)
        # Return a single string:
        return "\n".join(" " * indents + line for line in trimmed)


# Deprecated classes


class TemplateProposal(CodeAssistProposal):
    def __init__(self, name, template):
        warnings.warn(
            "TemplateProposal is deprecated.", DeprecationWarning, stacklevel=2
        )
        super().__init__(name, "template")
        self.template = template


class Template:
    def __init__(self, template):
        self.template = template
        warnings.warn("Template is deprecated.", DeprecationWarning, stacklevel=2)

    def variables(self):
        return []

    def substitute(self, mapping):
        return self.template

    def get_cursor_location(self, mapping):
        return len(self.template)



"""This module can be used for performing cross-project refactorings

See the "cross-project refactorings" section of ``docs/library.rst``
file.

"""

from rope.base import libutils, resources


class MultiProjectRefactoring:
    def __init__(self, refactoring, projects, addpath=True):
        """Create a multiproject proxy for the main refactoring

        `projects` are other project.

        """
        self.refactoring = refactoring
        self.projects = projects
        self.addpath = addpath

    def __call__(self, project, *args, **kwds):
        """Create the refactoring"""
        return _MultiRefactoring(
            self.refactoring, self.projects, self.addpath, project, *args, **kwds
        )


class _MultiRefactoring:
    def __init__(self, refactoring, other_projects, addpath, project, *args, **kwds):
        self.refactoring = refactoring
        self.projects = [project] + other_projects
        for other_project in other_projects:
            for folder in self.project.get_source_folders():
                other_project.get_prefs().add("python_path", folder.real_path)
        self.refactorings = []
        for other in self.projects:
            args, kwds = self._resources_for_args(other, args, kwds)
            self.refactorings.append(self.refactoring(other, *args, **kwds))

    def get_all_changes(self, *args, **kwds):
        """Get a project to changes dict"""
        result = []
        for project, refactoring in zip(self.projects, self.refactorings):
            args, kwds = self._resources_for_args(project, args, kwds)
            result.append((project, refactoring.get_changes(*args, **kwds)))
        return result

    def __getattr__(self, name):
        return getattr(self.main_refactoring, name)

    def _resources_for_args(self, project, args, kwds):
        newargs = [self._change_project_resource(project, arg) for arg in args]
        newkwds = {
            name: self._change_project_resource(project, value)
            for name, value in kwds.items()
        }
        return newargs, newkwds

    def _change_project_resource(self, project, obj):
        if isinstance(obj, resources.Resource) and obj.project != project:
            return libutils.path_to_resource(project, obj.real_path)
        return obj

    @property
    def project(self):
        return self.projects[0]

    @property
    def main_refactoring(self):
        return self.refactorings[0]


def perform(project_changes):
    for project, changes in project_changes:
        project.do(changes)
